

/*              RS232 Interface definition
 *
 *              Interrupt driven IO
 *
 */

#include <stdio.h>
#include <dos.h>
#include <int.h>        /* Symantec Interrupt Package */
#include "rs232.h"


#define INT_EOI         0x20    /* End Of Interrupt */
#define INT_CONTROLLER  0x021

#define COM1_INTVEC  0x0C               /* Vector number port1 */
#define COM2_INTVEC  0x0B               /* Vector number port2 */
#define BAUD_FREQ    1843200L           /* Frequenz Baudgenerator */


#define R_TXBUF         0x0
#define R_RXBUF         0x0
#define R_DIVL          0x0
#define R_DIVH          0x1
#define R_INTENABLE     0x1
#define R_INTID         0x2
#define R_LINECTRL      0x3
#define R_MODEMCTRL     0x4
#define R_LINESTAT      0x5
#define R_MODEMSTAT     0x6

#define OUT(reg, val) outp(com12[port]+reg,val)
#define IN(reg)       inp(com12[port]+reg)

static int com12[2] = { 0x3F8, 0x2F8 } ;
typedef unsigned char uchar ;
typedef unsigned int  uint ;
typedef enum {FALSE, TRUE} bool ;


#if !defined (__SC__)
/* Assembler supported functions */
extern int far int_port1() ;
extern int far int_port2() ;
#endif




struct RSBUF {
        uchar *rx_buf ;     /* pointer to receive_buffer */
        int   rx_size ;         /* size of receive_buffer    */
        int   rx_in   ;         /* index next char to read, used by driver */
        int   rx_out  ;         /* index char to read out, only if != rx_in */
        uchar *tx_buf ;    /* output spooler buffer  */
        int   tx_size ;
        int   tx_filled ;
        int   tx_in   ;
        int   tx_out  ;
        bool  tx_flag ;          /* synchronisation Spooler/Einfger */

        ModemStatus_t mstatus ;  /* modem-Status */
        uchar modemctrl;         /* shadow-register modem control */
  } ;

static
volatile struct RSBUF _rs_buf[2] ;   /* COM1 & COM 2  */

/*************** Prototypes *****************/
static void rs_error(void) ;


static
ModemStatus_t _near get_status(uint io_port)
{
   ModemStatus_t ms ;
   uchar         mmask ;

   mmask = inp(io_port) ;
   ms.CD   = (mmask & (1<<7)) != 0 ;
   ms.RING = (mmask & (1<<6)) != 0 ;
   ms.DSR  = (mmask & (1<<5)) != 0 ;
   ms.CTS  = (mmask & (1<<4)) != 0 ;
   return  ms ;
} /* get_status */


static
void  _near service_port(uint port_base, struct RSBUF *bp)
{
    unsigned  int_id ;
    uchar     ch ;

    for(;;) {
        int_id = inp(port_base+R_INTID) ;  /* Int Identification Register COM1 */
        if (int_id & 1) break ;  /* no more interrupts ?! */

        int_id >>= 1 ;  /* get interrupt source (lower 2 Bit */

        switch (int_id) {
        case 0:  /* MODEM Status Interrupt   */
           bp->mstatus = get_status(port_base+R_MODEMSTAT) ;
           break ;

        case 1:  /* transmitter hold empty interrupt */
           if (bp->tx_filled == 0)
              /* nothing more to spool */
              bp->tx_flag = TRUE ;
           else {
               ch = bp->tx_buf[bp->tx_out] ;
               if (++bp->tx_out >= bp->tx_size)
                  bp->tx_out = 0 ;
               bp->tx_filled-- ;
               outp (port_base+R_TXBUF, ch) ;
           }
           break ;

        case 2: /* receiver buffer , got a char */
           ch = inp(port_base+R_RXBUF) ;
           bp->rx_buf[bp->rx_in] = ch ;
           if (++bp->rx_in >= bp->rx_size)
               bp->rx_in = 0 ;
           break ;

        case 3:  /* Line Status Interrupt, jetzt noch nicht */
           break ;
        } /* switch */
   }/*for(;;) all Interrupts handled */

   outp (0x20, INT_EOI)  ; /* signal End Of Interrupt */
} /*service_port*/


static
int int_port1(struct INT_DATA *id)
{
   service_port(com12[0], &_rs_buf[0]) ;
   return 1 ;  /* no chain, return with IRET */
} /* int_port1 */


static
int int_port2(struct INT_DATA *id)
{
   service_port(com12[1], &_rs_buf[1]) ;
   return 1 ;  /* no chain, return with IRET */
} /* int_port2 */




/****************************************************************************
 *                                                                          *
 *             Initialize port with communication parameters needed         *
 *                                                                          *
 ****************************************************************************/

int rs_init (unsigned port, unsigned baud, unsigned data_bits,
             unsigned stop_bits, unsigned parity)
{
    unsigned divl=0, divh=0, ctr ;
    unsigned divisor ;

    if (port > 1)
       rs_error() ;

    divisor = (uint)((BAUD_FREQ / 16L)*10L / baud + 5L) / 10L ; /* mit Rundung */
    divh    = (divisor >> 8) ;
    divl    = divisor & 0xff ;

    OUT (R_LINECTRL, 0x80) ;  /* select divisor latch */
    OUT (R_DIVL, divl) ;
    OUT (R_DIVH, divh) ;

    OUT (R_LINECTRL, 0x00) ;  /* de-select divisor latch */
    ctr = (data_bits-5) |
          ((stop_bits == 2) ? 0x04 : 0) |
          ((parity != PNO) ? 0x08 : 0) |
          ((parity == PEVEN)  ? 0x10 : 0) ;

    OUT (R_LINECTRL, ctr) ;
    OUT (R_MODEMCTRL, 0x0B) ;  /* DTR, RTS to high */
    _rs_buf[port].modemctrl = 0x0B;
    return 0;
}/* end rs_init */




/****************************************************************************
 *                                                                          *
 *          Start driver by enabling interrupt used by seriell port         *
 *            set parameters as used by interrupt service routine           *                                                                          *
 *                                                                          *
 ****************************************************************************/

void rs_start(unsigned port, unsigned char *rx_adr, unsigned rx_size,
                             unsigned char *tx_adr, unsigned tx_size)
{
    struct RSBUF *rs_point ;
    int    cbit, status ;

    if (port > 1)
       rs_error() ;

    rs_point = &_rs_buf[port] ;
    rs_point->rx_size = rx_size ;
    rs_point->rx_buf  = rx_adr ;
    rs_point->rx_in   = 0 ;
    rs_point->rx_out  = 0 ;

    rs_point->tx_size   = tx_size ;
    rs_point->tx_buf    = tx_adr ;
    rs_point->tx_filled = 0 ;
    rs_point->tx_in     = 0 ;
    rs_point->tx_out    = 0 ;
    rs_point->tx_flag   = TRUE ;  /* 1. char send direct */
    rs_point->mstatus   = get_status(com12[port]+R_MODEMSTAT) ;

    if (port == COM1)
       status = int_intercept (COM1_INTVEC, int_port1, 0) ; /* use local stack */
    else
       status = int_intercept (COM2_INTVEC, int_port2, 0) ; /*  "" */

    if (status != 0) {
       fprintf (stderr, "\n--- ERROR intercepting INT-Vector ---\n") ;
    }

    if (rs_point->tx_size == 0)
       OUT (R_INTENABLE, 0x01|0x08) ;       /* interrupt on received data ready */
    else
       OUT (R_INTENABLE, 0x03|0x08) ;      /* INT on receive and Send Hold Empty */

    cbit = inp (INT_CONTROLLER) ;
    cbit &= ~(1 << (4-port)) ;
    outp (INT_CONTROLLER, cbit) ;     /* Enable hardware interrupt */


    /* Pending Interrupts lschen */
    for(;;) {
       int iid = IN(R_INTID) ;
       if((iid & 1) == 1) break  ;
       iid >>= 1 ;
       if(iid == 2)
           IN(R_RXBUF) ;  /* evtl. Mll lschen */
       if(iid == 0)
           IN(R_MODEMSTAT) ;  /* evtl. Mll lschen */
    }
}  /* end rs_start */




/****************************************************************************
 *                                                                          *
 *             Stop driver by disabling interrupt used by seriell port      *
 *                                                                          *
 ****************************************************************************/

int rs_stop(unsigned port)
{
    int cbit ;

    if (port > 1)
       rs_error() ;

    cbit = inp (INT_CONTROLLER) ; /* get interrupt mask */
    cbit |= (1 << (4-port)) ;     /* set interrupt bit used by port to 1*/
    outp (INT_CONTROLLER, cbit) ; /* restore interrupt mask */
    if (port == COM1)
       int_restore (COM1_INTVEC) ;
    else
       int_restore (COM2_INTVEC) ;

    return 0;
} /* end rs_stop */




/****************************************************************************
 *                                                                          *
 *                 Send a single character to seriell port                  *
 *                         non interrupt driven                             *
 *                                                                          *
 ****************************************************************************/

void rs_outchar (unsigned port, uchar ch)
/*  port :        port to be used */
/*    ch :        char to transmit */
{
    int st_bits ;
    volatile struct RSBUF *bp ;

    if (port > 1)
       rs_error() ;

    bp = &_rs_buf[port] ;
    if (bp->tx_size == 0) {
        /* output with polling */
        do {
           st_bits = IN(R_LINESTAT) ;
        } while ((st_bits & (1<<5)) == 0 ) ; /* Trans.Buf Empty */
        OUT(R_TXBUF,ch) ;
    }
    else {   /* interrupt spooled output */
        while (bp->tx_filled >= bp->tx_size) ; /* wait, spool buffer is full */
        int_off() ; /* no Interrupt during buffer manipulation */
        if (bp->tx_flag) {
            /* transmitter hold empty, send 1. char direct */
            bp->tx_flag = FALSE ;
            OUT(R_TXBUF, ch) ;
        }
        else {
            bp->tx_buf[bp->tx_in] = ch  ;
            if (++bp->tx_in >= bp->tx_size)
               bp->tx_in = 0 ;
            ++bp->tx_filled ;
        }
        int_on() ;  /* interrupts allowed again */
    } /* if else */
} /* end rs_outchar */




/****************************************************************************
 *                                                                          *
 *                   Get next character stored in Buffer                    *
 *                                                                          *
 *     Return value: char if character available, -1 if nothing available   *
 *                                                                          *
 ****************************************************************************/

int rs_inchar(unsigned port)
/* port  :  port to be used */
{
    volatile struct RSBUF *rs_point ;
    int  readchar ;

    if (port > 1)
       rs_error() ;

    rs_point = &_rs_buf[port] ;
    if ( rs_point->rx_in != rs_point->rx_out ) {
        readchar = rs_point->rx_buf[rs_point->rx_out] ;
        readchar &= 0xff ; /* make shure it is unsigned extended */
        if (++rs_point->rx_out >= rs_point->rx_size)
            rs_point->rx_out = 0 ;
        return (readchar) ;
    }

    return -1 ;  /* nothing available yet */
} /* end rs_inchar */



/****************************************************************************
 *                                                                          *
 *            Send a whole string (terminated by NULL-Charakter)            *
 *                         non interrupt driven                             *
 *                                                                          *
 ****************************************************************************/

void rs_sendline(unsigned port, char *line)
/*  port :           port to use */
/*  line :           pointer to line to be send */
{
    if (port > 1)
       rs_error() ;

    while (*line != 0)
          rs_outchar(port, (uchar)*line++) ;
}  /* rs_sendline */


void rs_sendblock(unsigned port, void *block, unsigned bsize)
/*  port  :     port to use */
/*  block :     pointer to mem-block to be send */
/*  bsize :     blocksize  */
{
    uchar *bl ;

    if (port > 1)
       rs_error() ;

    bl = (uchar*)block ;
    while (bsize-- > 0)
          rs_outchar(port, *bl++) ;
}  /* rs_sendblock */




/****************************************************************************
 *                                                                          *
 *           Return number of bytes availble in receive buffer              *
 *                                                                          *
 ****************************************************************************/

int rs_status(unsigned port)
{
    volatile struct RSBUF *rs_point ;
    int available ;

    if (port > 1)
       rs_error() ;

    rs_point = &_rs_buf[port] ;

    if (rs_point->rx_in < rs_point->rx_out)
       available = rs_point->rx_size - rs_point->rx_out + rs_point->rx_in ;
    else
       available = rs_point->rx_in - rs_point->rx_out ;

    return (available) ;
} /* rs_status */


int rs_remaining(unsigned port)
{
    if (port > 1)
       rs_error() ;

    return _rs_buf[port].tx_filled ;
} /* rs_remaining */


ModemStatus_t rs_modemstatus(unsigned port)
{
    if (port > 1)
       rs_error() ;

    return _rs_buf[port].mstatus ;
} /* rs_modemstatus */


void rs_setdtr(unsigned port, int v)  /* v=0 -> low pegel, v!=0 -> high */
{
  if(port < 2) {
    struct RSBUF *rs = &_rs_buf[port] ;
    if(v)
      rs->modemctrl |= 1;  /* DTR-Bit 0 setzen */
    else
      rs->modemctrl &= ~1; /* DTR-Bit 0 lschen */
    OUT(R_MODEMCTRL, rs->modemctrl);
  }
}/*rs_setdtr*/


void rs_setrts(unsigned port, int v)  /* v=0 -> low pegel, v!=0 -> high */
{
  if(port < 2) {
    struct RSBUF *rs = &_rs_buf[port] ;
    if(v)
      rs->modemctrl |= 2;  /* RTS-Bit 1 setzen */
    else
      rs->modemctrl &= ~2; /* RTS-Bit 1 lschen */
    OUT(R_MODEMCTRL, rs->modemctrl);
  }
}/*rs_setrts*/


/****************************************************************************
 *                                                                          *
 *               Error message if portnumber not in valid range             *
 *                                                                          *
 ****************************************************************************/

static
void rs_error(void)
{
     fprintf(stderr,"\n--- FATAL ERROR: Invalid port used ---\n") ;
     exit(1) ;
} /* rs_error */


#ifdef TEST
static int delay (ticks)
int ticks ;
{
        int     sink ;

        while (ticks-- > 0)
                sink = ticks * ticks ;
} /* delay */

#endif

